/*******************************************************************************
 * GenRGenS v.1.0, Generation of Random Genomic Sequences                      *
 * Copyright (C) 2001  Alain Denise/Yann Ponty                                 *
 *                                                                             *
 * This program is free software; you can redistribute it and/or               *
 * modify it under the terms of the GNU General Public License                 *
 * as published by the Free Software Foundation; either version 2              *
 * of the License, or (at your option) any later version.                      *
 *                                                                             *
 * This program is distributed in the hope that it will be useful,             *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 * GNU General Public License for more details.                                *
 *                                                                             *
 * You should have received a copy of the GNU General Public License           *
 * along with this program; if not, write to the Free Software                 *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *******************************************************************************/

package GenRGenS.grammar;

import java.util.Vector;

/**
 * classe ContextFreeGrammarProduction 
 *     contient a la fois partie gauche et partie droite         
 *     le nombre de lettre est fix a la cration                
 *     une reference hors des limites provoque une exception
 */


public class ContextFreeGrammarProduction
{
		private int[] lettre;

    /**
     * Construit une instance de <code>ContextFreeGrammarProduction</code> de taille fixe.
     * Elle modlise une rgle de la grammaire.
     * @param nombreDeLettres Le nombres de symboles de la rgle (= NBRgle de la partie droite + 1)
     */
		ContextFreeGrammarProduction(int nombreDeLettres)
		{
				lettre = new int[nombreDeLettres];
		}

		
		/** 
		 * Construit une instance de <code>ContextFreeGrammarProduction</code> de contenu fix.
     * Elle modlise une rgle de la grammaire.
     * @param symboles Un vecteur gnrique ne contenant que des instances de <code>Integer</code> 
     * dont les contenus entiers sont les index des symbole de la rgle.
		 */
		ContextFreeGrammarProduction(Vector symboles)
				throws BadGrammarException
		{
				if (symboles.isEmpty())
						throw(new BadGrammarException("incorrect Vector for ContextFreeGrammarProduction construction"));
				int longueur;
				lettre = new int[longueur=symboles.size()];
				
				for (int i=0;i<longueur;i++)
						lettre[i]=((Integer) symboles.get(i)).intValue();						
		}

    /**
     * Modifie le contenu d'une rgle.
     * Substitue  un symbole prsent un autre symbole dcrit par son index.
     * @param index L'emplacement du symbole  modifier.
     * @param valeur L'index du symbole  lui substituer.
     */

		void put(int index, int valeur)
		{
				if (index<lettre.length)
						lettre[index]=valeur;
				else
						throw( new IndexOutOfBoundsException());
		}
		
		/**
		 * Renvoie l'index d'un symbole de la rgle selon sa position.
		 * @param index L'emplacement dans la rgle du symbole.
		 * @return L'index du symbole vis.
		 */

		int get(int index)
		{
				if (index<lettre.length)
						return(lettre[index]);
				else
						throw( new IndexOutOfBoundsException());
		}

		/**
		 * Renvoie la partie gauche de la rgle.
		 * @return L'index de la partie gauche de la rgle.
		 */
		int getLHS()
		{ return(lettre[0]); };
		
		/**
		 * Renvoie le nombre de symbole dans cette rgle.
		 * @return Le nombre de symbole dans cette rgle.
		 */
		int size()
		{	return(lettre.length); };

		/**
		 * Le langage associ  cette rgle est il rduit  epsilon ?
		 * @return true si la rgle ne drive qu'epsilon, false sinon.
		 */
		boolean isEpsilonProduction()
		{	return(lettre.length==1);	};

		/**
		 * Le langage associ  cette rgle est il rduit  un symbole terminal ?
		 * @return true si la rgle ne drive qu'un terminal, false sinon.
		 */
		boolean isTerminalProduction()
		{ return(lettre.length==2 && lettre[1]<0); };

		/**
		 * Le langage associ  cette rgle est il rduit  celui d'un symbole non-terminal ?
		 * @return true si la rgle ne drive qu'un non-terminal, false sinon.
		 */
		boolean isSingleProduction()
		{ return(lettre.length==2 && lettre[1]>=0);	};
		
		/**
		 * Renvoie une reprsentation littrale de cette instance.
		 */
		String print()
		{
				String temp="";
				for (int i=0;i<lettre.length;i++)	temp+=lettre[i]+" ";
				return(temp);
		}

}
